@@ -12,6 +12,7 @@
 #include <ruby/st.h>
 #include <ruby/io.h>
 #include <ruby/intern.h>
+#include <ruby/vm.h>
 #include <signal.h>
 #include <sys/time.h>
 #include <time.h>
@@ -32,6 +33,7 @@ static const char *fake_frame_cstrs[] = {
 };
 
 static int stackprof_use_postponed_job = 1;
+static int ruby_vm_running = 0;
 
 #define TOTAL_FAKE_FRAMES (sizeof(fake_frame_cstrs) / sizeof(char *))
 
@@ -125,6 +127,8 @@ static struct {
     sample_time_t buffer_time;
     VALUE frames_buffer[BUF_SIZE];
     int lines_buffer[BUF_SIZE];
+
+    pthread_t target_thread;
 } _stackprof;
 
 static VALUE sym_object, sym_wall, sym_cpu, sym_custom, sym_name, sym_file, sym_line;
@@ -219,6 +223,7 @@ stackprof_start(int argc, VALUE *argv, VALUE self)
     _stackprof.ignore_gc = ignore_gc;
     _stackprof.metadata = metadata;
     _stackprof.out = out;
+    _stackprof.target_thread = pthread_self();
 
     if (raw) {
 	capture_timestamp(&_stackprof.last_sample_at);
@@ -721,7 +726,27 @@ stackprof_signal_handler(int sig, siginfo_t *sinfo, void *ucontext)
     _stackprof.overall_signals++;
 
     if (!_stackprof.running) return;
-    if (!ruby_native_thread_p()) return;
+
+    // There's a possibility that the signal handler is invoked *after* the Ruby
+    // VM has been shut down (e.g. after ruby_cleanup(0)). In this case, things
+    // that rely on global VM state (e.g. rb_during_gc) will segfault.
+    if (!ruby_vm_running) return;
+
+    if (_stackprof.mode == sym_wall) {
+        // In "wall" mode, the SIGALRM signal will arrive at an arbitrary thread.
+        // In order to provide more useful results, especially under threaded web
+        // servers, we want to forward this signal to the original thread
+        // StackProf was started from.
+        // According to POSIX.1-2008 TC1 pthread_kill and pthread_self should be
+        // async-signal-safe.
+        if (pthread_self() != _stackprof.target_thread) {
+            pthread_kill(_stackprof.target_thread, sig);
+            return;
+        }
+    } else {
+        if (!ruby_native_thread_p()) return;
+    }
+
     if (pthread_mutex_trylock(&lock)) return;
 
     if (!_stackprof.ignore_gc && rb_during_gc()) {
@@ -786,6 +811,10 @@ stackprof_gc_mark(void *data)
 
     if (_stackprof.frames)
 	st_foreach(_stackprof.frames, frame_mark_i, 0);
+
+    for (int i = 0; i < _stackprof.buffer_count; i++) {
+        rb_gc_mark(_stackprof.frames_buffer[i]);
+    }
 }
 
 static void
@@ -827,6 +856,12 @@ stackprof_use_postponed_job_l(VALUE self)
     return Qnil;
 }
 
+static void
+stackprof_at_exit(ruby_vm_t* vm)
+{
+    ruby_vm_running = 0;
+}
+
 void
 Init_stackprof(void)
 {
@@ -837,6 +872,9 @@ Init_stackprof(void)
     */
     stackprof_use_postponed_job = RUBY_API_VERSION_MAJOR < 3;
 
+    ruby_vm_running = 1;
+    ruby_vm_at_exit(stackprof_at_exit);
+
 #define S(name) sym_##name = ID2SYM(rb_intern(#name));
     S(object);
     S(custom);
